from typing import Optional, Union, List, Dict
from io import BytesIO
from json import loads

import numpy as np

from litecow.common.litecow_pb2 import InferenceRequest
from litecow.common.litecow_pb2_grpc import ICOWStub
from litecow.common import common


class ICOWClient:
    def __init__(self, channel):
        """Initializes an ICOWClient

        Parameters
        ----------
        channel : grpc._channel.Channel
            The channel to connect with the server on

        Examples
        --------
        >>> ICOWClient(grpc.insecure_channel('localhost:8080'))
        """
        self.stub = ICOWStub(channel)

    def get_inference(
        self,
        model_key: str,
        inputs: Union[Dict[str, np.ndarray], List[np.ndarray], np.ndarray],
        outputs: Optional[List[str]] = None,
        model_version: Optional[str] = None,
    ) -> Dict[str, np.ndarray]:
        """ Get inference with input arrays

        Parameters
        ----------
        model_key : str
            The key to the model in s3.
        inputs : Union[Dict[str, np.ndarray], List[np.ndarray], np.ndarray]
            The input(s) to run inference on may be a dict 
            mapping input names to arrays, a list of input arrays, or a single input array
        outputs : Optional[List[str]] = None
            The outputs of the inference to return, if
            None all will be returned.
        model_version : Optional[str]
            The version of the model to use.

        Returns
        -------
        Dict[str, np.ndarray]
            Results of inference

        Examples
        --------
        >>> client.get_inference("s3://models/my_model_key.onnx", np.float32(np.random.random((1, 3, 10, 10))), outputs=["loc", "landms"], model_version="v1")
        """
        # ensure inputs is formatted correctly
        if isinstance(inputs, Dict):
            named_inputs = common.prepare_named_arrays(inputs)
        elif isinstance(inputs, List):
            unnamed_inputs = common.prepare_array_list(inputs)
        else:
            unnamed_inputs = common.prepare_array_list([inputs])
        request_kwargs = locals()

        field_names = InferenceRequest.DESCRIPTOR.fields_by_name.keys()
        filtered_kwargs = {
            key: value
            for key, value in request_kwargs.items()
            if key in field_names and value
        }
        return common.unprepare_named_arrays(
            self.stub.get_inference(InferenceRequest(**filtered_kwargs))
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: litecow/common/litecow.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='litecow/common/litecow.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1clitecow/common/litecow.proto\"\xa3\x01\n\x10InferenceRequest\x12\x11\n\tmodel_key\x18\x01 \x01(\t\x12\x15\n\rmodel_version\x18\x02 \x01(\t\x12$\n\x0cnamed_inputs\x18\x03 \x01(\x0b\x32\x0c.NamedArraysH\x00\x12$\n\x0eunnamed_inputs\x18\x04 \x01(\x0b\x32\n.ArrayListH\x00\x12\x0f\n\x07outputs\x18\x05 \x03(\tB\x08\n\x06inputs\"\x7f\n\x0bNamedArrays\x12\x34\n\rname_to_array\x18\x01 \x03(\x0b\x32\x1d.NamedArrays.NameToArrayEntry\x1a:\n\x10NameToArrayEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x15\n\x05value\x18\x02 \x01(\x0b\x32\x06.Array:\x02\x38\x01\"#\n\tArrayList\x12\x16\n\x06\x61rrays\x18\x01 \x03(\x0b\x32\x06.Array\"\xb1\x01\n\x05\x41rray\x12\r\n\x05shape\x18\x01 \x03(\r\x12\"\n\x0b\x66loat_array\x18\x02 \x01(\x0b\x32\x0b.FloatArrayH\x00\x12$\n\x0c\x64ouble_array\x18\x03 \x01(\x0b\x32\x0c.DoubleArrayH\x00\x12\"\n\x0bint32_array\x18\x04 \x01(\x0b\x32\x0b.Int32ArrayH\x00\x12\"\n\x0bint64_array\x18\x05 \x01(\x0b\x32\x0b.Int64ArrayH\x00\x42\x07\n\x05\x61rray\"\x1c\n\nInt32Array\x12\x0e\n\x06values\x18\x01 \x03(\x05\"\x1c\n\nInt64Array\x12\x0e\n\x06values\x18\x01 \x03(\x03\"\x1c\n\nFloatArray\x12\x0e\n\x06values\x18\x01 \x03(\x02\"\x1d\n\x0b\x44oubleArray\x12\x0e\n\x06values\x18\x01 \x03(\x01\x32:\n\x04ICOW\x12\x32\n\rget_inference\x12\x11.InferenceRequest\x1a\x0c.NamedArrays\"\x00\x62\x06proto3'
)




_INFERENCEREQUEST = _descriptor.Descriptor(
  name='InferenceRequest',
  full_name='InferenceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_key', full_name='InferenceRequest.model_key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_version', full_name='InferenceRequest.model_version', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='named_inputs', full_name='InferenceRequest.named_inputs', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='unnamed_inputs', full_name='InferenceRequest.unnamed_inputs', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='outputs', full_name='InferenceRequest.outputs', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='inputs', full_name='InferenceRequest.inputs',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=33,
  serialized_end=196,
)


_NAMEDARRAYS_NAMETOARRAYENTRY = _descriptor.Descriptor(
  name='NameToArrayEntry',
  full_name='NamedArrays.NameToArrayEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='NamedArrays.NameToArrayEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='NamedArrays.NameToArrayEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=267,
  serialized_end=325,
)

_NAMEDARRAYS = _descriptor.Descriptor(
  name='NamedArrays',
  full_name='NamedArrays',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name_to_array', full_name='NamedArrays.name_to_array', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_NAMEDARRAYS_NAMETOARRAYENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=198,
  serialized_end=325,
)


_ARRAYLIST = _descriptor.Descriptor(
  name='ArrayList',
  full_name='ArrayList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='arrays', full_name='ArrayList.arrays', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=327,
  serialized_end=362,
)


_ARRAY = _descriptor.Descriptor(
  name='Array',
  full_name='Array',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='shape', full_name='Array.shape', index=0,
      number=1, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='float_array', full_name='Array.float_array', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='double_array', full_name='Array.double_array', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='int32_array', full_name='Array.int32_array', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='int64_array', full_name='Array.int64_array', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='array', full_name='Array.array',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=365,
  serialized_end=542,
)


_INT32ARRAY = _descriptor.Descriptor(
  name='Int32Array',
  full_name='Int32Array',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='values', full_name='Int32Array.values', index=0,
      number=1, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=544,
  serialized_end=572,
)


_INT64ARRAY = _descriptor.Descriptor(
  name='Int64Array',
  full_name='Int64Array',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='values', full_name='Int64Array.values', index=0,
      number=1, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=574,
  serialized_end=602,
)


_FLOATARRAY = _descriptor.Descriptor(
  name='FloatArray',
  full_name='FloatArray',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='values', full_name='FloatArray.values', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=604,
  serialized_end=632,
)


_DOUBLEARRAY = _descriptor.Descriptor(
  name='DoubleArray',
  full_name='DoubleArray',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='values', full_name='DoubleArray.values', index=0,
      number=1, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=634,
  serialized_end=663,
)

_INFERENCEREQUEST.fields_by_name['named_inputs'].message_type = _NAMEDARRAYS
_INFERENCEREQUEST.fields_by_name['unnamed_inputs'].message_type = _ARRAYLIST
_INFERENCEREQUEST.oneofs_by_name['inputs'].fields.append(
  _INFERENCEREQUEST.fields_by_name['named_inputs'])
_INFERENCEREQUEST.fields_by_name['named_inputs'].containing_oneof = _INFERENCEREQUEST.oneofs_by_name['inputs']
_INFERENCEREQUEST.oneofs_by_name['inputs'].fields.append(
  _INFERENCEREQUEST.fields_by_name['unnamed_inputs'])
_INFERENCEREQUEST.fields_by_name['unnamed_inputs'].containing_oneof = _INFERENCEREQUEST.oneofs_by_name['inputs']
_NAMEDARRAYS_NAMETOARRAYENTRY.fields_by_name['value'].message_type = _ARRAY
_NAMEDARRAYS_NAMETOARRAYENTRY.containing_type = _NAMEDARRAYS
_NAMEDARRAYS.fields_by_name['name_to_array'].message_type = _NAMEDARRAYS_NAMETOARRAYENTRY
_ARRAYLIST.fields_by_name['arrays'].message_type = _ARRAY
_ARRAY.fields_by_name['float_array'].message_type = _FLOATARRAY
_ARRAY.fields_by_name['double_array'].message_type = _DOUBLEARRAY
_ARRAY.fields_by_name['int32_array'].message_type = _INT32ARRAY
_ARRAY.fields_by_name['int64_array'].message_type = _INT64ARRAY
_ARRAY.oneofs_by_name['array'].fields.append(
  _ARRAY.fields_by_name['float_array'])
_ARRAY.fields_by_name['float_array'].containing_oneof = _ARRAY.oneofs_by_name['array']
_ARRAY.oneofs_by_name['array'].fields.append(
  _ARRAY.fields_by_name['double_array'])
_ARRAY.fields_by_name['double_array'].containing_oneof = _ARRAY.oneofs_by_name['array']
_ARRAY.oneofs_by_name['array'].fields.append(
  _ARRAY.fields_by_name['int32_array'])
_ARRAY.fields_by_name['int32_array'].containing_oneof = _ARRAY.oneofs_by_name['array']
_ARRAY.oneofs_by_name['array'].fields.append(
  _ARRAY.fields_by_name['int64_array'])
_ARRAY.fields_by_name['int64_array'].containing_oneof = _ARRAY.oneofs_by_name['array']
DESCRIPTOR.message_types_by_name['InferenceRequest'] = _INFERENCEREQUEST
DESCRIPTOR.message_types_by_name['NamedArrays'] = _NAMEDARRAYS
DESCRIPTOR.message_types_by_name['ArrayList'] = _ARRAYLIST
DESCRIPTOR.message_types_by_name['Array'] = _ARRAY
DESCRIPTOR.message_types_by_name['Int32Array'] = _INT32ARRAY
DESCRIPTOR.message_types_by_name['Int64Array'] = _INT64ARRAY
DESCRIPTOR.message_types_by_name['FloatArray'] = _FLOATARRAY
DESCRIPTOR.message_types_by_name['DoubleArray'] = _DOUBLEARRAY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

InferenceRequest = _reflection.GeneratedProtocolMessageType('InferenceRequest', (_message.Message,), {
  'DESCRIPTOR' : _INFERENCEREQUEST,
  '__module__' : 'litecow.common.litecow_pb2'
  # @@protoc_insertion_point(class_scope:InferenceRequest)
  })
_sym_db.RegisterMessage(InferenceRequest)

NamedArrays = _reflection.GeneratedProtocolMessageType('NamedArrays', (_message.Message,), {

  'NameToArrayEntry' : _reflection.GeneratedProtocolMessageType('NameToArrayEntry', (_message.Message,), {
    'DESCRIPTOR' : _NAMEDARRAYS_NAMETOARRAYENTRY,
    '__module__' : 'litecow.common.litecow_pb2'
    # @@protoc_insertion_point(class_scope:NamedArrays.NameToArrayEntry)
    })
  ,
  'DESCRIPTOR' : _NAMEDARRAYS,
  '__module__' : 'litecow.common.litecow_pb2'
  # @@protoc_insertion_point(class_scope:NamedArrays)
  })
_sym_db.RegisterMessage(NamedArrays)
_sym_db.RegisterMessage(NamedArrays.NameToArrayEntry)

ArrayList = _reflection.GeneratedProtocolMessageType('ArrayList', (_message.Message,), {
  'DESCRIPTOR' : _ARRAYLIST,
  '__module__' : 'litecow.common.litecow_pb2'
  # @@protoc_insertion_point(class_scope:ArrayList)
  })
_sym_db.RegisterMessage(ArrayList)

Array = _reflection.GeneratedProtocolMessageType('Array', (_message.Message,), {
  'DESCRIPTOR' : _ARRAY,
  '__module__' : 'litecow.common.litecow_pb2'
  # @@protoc_insertion_point(class_scope:Array)
  })
_sym_db.RegisterMessage(Array)

Int32Array = _reflection.GeneratedProtocolMessageType('Int32Array', (_message.Message,), {
  'DESCRIPTOR' : _INT32ARRAY,
  '__module__' : 'litecow.common.litecow_pb2'
  # @@protoc_insertion_point(class_scope:Int32Array)
  })
_sym_db.RegisterMessage(Int32Array)

Int64Array = _reflection.GeneratedProtocolMessageType('Int64Array', (_message.Message,), {
  'DESCRIPTOR' : _INT64ARRAY,
  '__module__' : 'litecow.common.litecow_pb2'
  # @@protoc_insertion_point(class_scope:Int64Array)
  })
_sym_db.RegisterMessage(Int64Array)

FloatArray = _reflection.GeneratedProtocolMessageType('FloatArray', (_message.Message,), {
  'DESCRIPTOR' : _FLOATARRAY,
  '__module__' : 'litecow.common.litecow_pb2'
  # @@protoc_insertion_point(class_scope:FloatArray)
  })
_sym_db.RegisterMessage(FloatArray)

DoubleArray = _reflection.GeneratedProtocolMessageType('DoubleArray', (_message.Message,), {
  'DESCRIPTOR' : _DOUBLEARRAY,
  '__module__' : 'litecow.common.litecow_pb2'
  # @@protoc_insertion_point(class_scope:DoubleArray)
  })
_sym_db.RegisterMessage(DoubleArray)


_NAMEDARRAYS_NAMETOARRAYENTRY._options = None

_ICOW = _descriptor.ServiceDescriptor(
  name='ICOW',
  full_name='ICOW',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=665,
  serialized_end=723,
  methods=[
  _descriptor.MethodDescriptor(
    name='get_inference',
    full_name='ICOW.get_inference',
    index=0,
    containing_service=None,
    input_type=_INFERENCEREQUEST,
    output_type=_NAMEDARRAYS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_ICOW)

DESCRIPTOR.services_by_name['ICOW'] = _ICOW

# @@protoc_insertion_point(module_scope)



from tempfile import NamedTemporaryFile
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, Tuple, Dict
from argparse import ArgumentParser
from os import environ
from urllib.parse import urlparse

import boto3
import grpc
from onnxruntime import InferenceSession
import numpy as np

from litecow.common.litecow_pb2_grpc import ICOWServicer, add_ICOWServicer_to_server
from litecow.common import common


class ICOWServicer(ICOWServicer):
    def __init__(
        self,
        batch_hint: int,
        aws_access_key: str,
        aws_secret_key: str,
        endpoint_url: str,
    ):
        """Initializes ICOWService

        Parameters
        ----------
        batch_hint: int
            Size to start batches at. Automatically reduced when overflow occurs.
        aws_access_key: string
            Access key for s3 storage
        aws_secret_key: string
            Secret key for s3 stroage
        endpoint_url: string
            Url for s3 storage
        """
        self.batch_hint = batch_hint
        self.s3_client = boto3.client(
            "s3",
            aws_access_key_id=aws_access_key,
            aws_secret_access_key=aws_secret_key,
            endpoint_url=endpoint_url,
        )

    def get_inference(self, request, context):
        """Handles requests for service get_inference calls.

        Parameters
        ----------
        request: litecow_common.litecow_pb2.InferenceRequest
            The InferenceRequest to answer.
        context: grpc.ServicerContext
            The context for this request

        Returns
        -------
        Union[litecow_common.litecow_pb2.NamedArrays, litecow_common.litecow_pb2.ArrayList]
            Outputs of inference encoded as a NamedArrays
        """
        # first retrieve model
        try:
            model = self._get_model_version_from_s3(
                request.model_key, request.model_version
            )
        except Exception as error:
            context.set_code(grpc.StatusCode.FAILED_PRECONDITION)
            context.set_details(f"Error occured while loading model from s3:\n{error}")
            raise error

        # get inputs ready for model
        if request.HasField('named_inputs'):
            inputs = common.unprepare_named_arrays(request.named_inputs)
        elif request.HasField('unnamed_inputs'):
            inputs = common.unprepare_array_list(request.unnamed_inputs)
            inputs = {model_input.name:array for model_input, array in zip(model.get_inputs(), inputs)}
        else:
            context.set_code(grpc.StatusCode.INVALID_ARGUMENT)
            context.set_details("Client provided request without inputs")
            raise ValueError("Request must have value for named_inputs or unnamed_inputs")

        outputs = request.outputs
        try:
            result = model.run(outputs, inputs)
        except Exception as error:
            context.set_code(grpc.StatusCode.UNKNOWN)
            context.set_details(f"Error occured while doing inference:\n{error}")
            raise error

        # label and prepare result to be sent
        output_labels = map(lambda x: x.name, model.get_outputs())
        output_labels = filter(lambda x: x in outputs, output_labels) if outputs else output_labels
        labeled_result = {name:array for name,array in zip(output_labels, result)}

        return common.prepare_named_arrays(labeled_result)

    def _get_model_version_from_s3(
        self, s3_path: str, model_version: Optional[str]
    ) -> InferenceSession:
        """Retrieves a model version from s3

        Parameters
        ----------
        s3_path: str
            s3 object key for model to load should use s3 scheme like
            s3://models/facenet.onnx
        model_version: Optional[str]
            The optional version of the model to load
            empty is treated like None.

        Returns
        -------
        onnxruntime.InferenceSession:
            The requested model loaded into an inference session
        """
        parsed_url = urlparse(s3_path)
        bucket_name, object_name = parsed_url.netloc, parsed_url.path

        if parsed_url.scheme != "s3":
            raise Exception("Model s3_path must follow s3 protocol")
        if not model_version:
            model_version = None

        def is_correct_version(obj_version) -> Optional[Tuple[str, dict]]:
            version_id = obj_version["VersionId"]
            # check if any of the object_tags contain {"model-version":model_version}
            object_tagging = self.s3_client.get_object_tagging(Bucket=bucket_name, Key=object_name, VersionId=version_id)["TagSet"]
            has_right_version = any(map(lambda tag: tag["Key"] == "model-version" and tag["Value"] == model_version, object_tagging))
            return (version_id, object_tagging) if has_right_version else None

        # resolve version of obj
        if model_version:
            try:
                correct_version, obj_meta = next(
                    filter(
                        None,
                        map(
                            is_correct_version,
                            self.s3_client.list_object_versions(
                                Bucket=bucket_name, Prefix=object_name
                            )["Versions"],
                        ),
                    )
                )
            except Exception as error:
                raise Exception(
                    f"Error occured while checking for object version:\n{error}"
                )

        # retrieve the object
        with NamedTemporaryFile() as temp_file:
            self.s3_client.download_fileobj(
                bucket_name,
                object_name,
                temp_file,
                ExtraArgs={"VersionId": correct_version} if model_version else None,
            )
            return InferenceSession(temp_file.name)


def serve(
    workers: int,
    port: int,
    batch_hint: int,
    aws_access_key: str,
    aws_secret_key: str,
    aws_endpoint_url: str,
) -> None:
    """Run the ICOWServicer with arguments passed.

    Parameters
    ----------
    workers: int
        The number of max workers for the ThreadPoolExecutor running the server.
    port: int
        The port number to bind the service to.
    batch_hint: int
        The initial number of inputs to batch together in a single forward pass.
    aws_access_key: str
        The aws access key for the s3 storage to be used by the service.
    aws_secret_key: str
        The aws secret key for the s3 storage to be used by the service.
    aws_endpoint_url: str
        The url for the s3 storage used by the service includes scheme and port.
        Like http://my-s3-service.us:1337
    """
    server = grpc.server(ThreadPoolExecutor(max_workers=workers))
    add_ICOWServicer_to_server(
        ICOWServicer(batch_hint, aws_access_key, aws_secret_key, aws_endpoint_url),
        server,
    )
    server.add_insecure_port(f"0.0.0.0:{port}")
    server.start()
    server.wait_for_termination()


def main():
    """Run the icow_service with the arguments provided on the command line.

    Uses arguments provided on the command line and environment variables set on
    the system to run an icow_service. Must set environment variables: AWS_ACCESS_KEY
    AWS_SECRET_KEY, S3ENDPOINT_URL to values for the s3 storage so the icow_service
    can access s3.
    """
    parser = ArgumentParser(description="Run GRPC server for inference.")
    parser.add_argument(
        "num_worker", type=int, help="Number of thread workers for server."
    )
    parser.add_argument("port", type=int, help="Port number to serve on.")
    parser.add_argument(
        "batch_hint",
        type=int,
        nargs="?",
        default=16,
        help="Batch size inference should start with. Higher values may overflow the gpu or system memory and cause multiple retries to be necessary for an inference to complete, as the batch size is reduce by at each failure.",
    )
    args = parser.parse_args()
    serve(
        args.num_worker,
        args.port,
        args.batch_hint,
        environ["AWS_ACCESS_KEY"],
        environ["AWS_SECRET_KEY"],
        environ["S3ENDPOINT_URL"],
    )


if __name__ == "__main__":
    main()



from typing import Dict, List

import numpy as np

from . import litecow_pb2


def prepare_array(array: np.ndarray) -> litecow_pb2.Array:
    """Convert a numpy array to a litecow array

    Parameters
    ----------
    array: np.ndarray
        Array to convert

    Returns
    -------
    litecow.common.litecow_pb2.Array
    """
    array_type = str(array.dtype)
    array_cls = {
        "float64": ("double_array", litecow_pb2.DoubleArray),
        "float32": ("float_array", litecow_pb2.FloatArray),
        "int32": ("int32_array", litecow_pb2.Int32Array),
        "int64": ("int64_array", litecow_pb2.Int64Array),
    }
    f_name, cls = array_cls[array_type]
    kwargs = {"shape": array.shape, f_name: cls(values=array.flatten().tolist())}
    return litecow_pb2.Array(**kwargs)


def prepare_named_arrays(
    named_arrays: Dict[str, np.ndarray]
) -> litecow_pb2.NamedArrays:
    """Convert a dictionary of arrays to a litecow NamedArrays

    Parameters
    ----------
    named_arrays: Dict[str, np.ndarray]
        Dictionary mapping string names to numpy arrays

    Returns
    -------
    litecow.common.litecow_pb2.NamedArrays
    """
    return litecow_pb2.NamedArrays(
        name_to_array={
            name: prepare_array(array) for name, array in named_arrays.items()
        }
    )


def prepare_array_list(arrays: List[np.ndarray]) -> litecow_pb2.ArrayList:
    """Convert a list of numpy arrays to a litecow ArrayList

    Parameters
    ----------
    arrays: List[np.ndarray]
        Arrays to convert to litecow ArrayList

    Returns
    -------
    litecow.common.litecow_pb2.ArrayList
    """
    return litecow_pb2.ArrayList(arrays=list(map(prepare_array, arrays)))

def unprepare_array(array:  litecow_pb2.Array) -> np.ndarray:
    """Convert a litecow Array to a numpy array

    Parameters
    ----------
    array: litecow.common.litecow_pb2.Array
        Array to convert back to numpy array

    Returns
    -------
    np.ndarray
    """
    if array.HasField('double_array'):
        np_array = np.float64(array.double_array.values)
    elif array.HasField('float_array'):
        np_array = np.float32(array.float_array.values)
    elif array.HasField('int32_array'):
        np_array = np.int32(array.int32_array.values)
    elif array.HasField('int64_array'):
        np_array = np.int64(array.int64_array.values)
    return np_array.reshape(array.shape)

def unprepare_named_arrays(named_arrays: litecow_pb2.NamedArrays) -> Dict[str, np.ndarray]:
    """Convert litecow NamedArrays back to a dictionary of numpy arrays

    Parameters
    ----------
    named_arrays: litecow.common.litecow_pb2.NamedArrays
        NamedArrays to convert

    Returns
    -------
    Dict[str, np.ndarray]
    """
    return {name:unprepare_array(array) for name, array in named_arrays.name_to_array.items()}

def unprepare_array_list(arrays: litecow_pb2.ArrayList) -> List[np.ndarray]:
    """Convert litecow ArrayList back to a list of numpy arrays

    Parameters
    ----------
    arrays: litecow.common.litecow_pb2.ArrayList
        ArrayList to convert.

    Returns
    -------
    List[np.ndarray]
    """
    return list(map(unprepare_array, arrays.arrays))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['litecow', 'litecow.common']

package_data = \
{'': ['*']}

install_requires = \
['grpcio>=1.33.2,<2.0.0', 'numpy>=1.17.4,<2.0.0', 'protobuf']

extras_require = \
{'server-cpu': ['onnxruntime>=1.7.0,<2.0.0', 'boto3>=1.11.14,<2.0.0'],
 'server-gpu': ['onnxruntime-gpu>=1.7.0,<2.0.0', 'boto3>=1.11.14,<2.0.0']}

entry_points = \
{'console_scripts': ['litecow-service = litecow.server:main']}

setup_kwargs = {
    'name': 'litecow',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': 'Striveworks',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
